package edu.upvehu.gbg.docencia.perfectos;

import java.math.BigInteger;
import java.util.concurrent.ArrayBlockingQueue;
import java.util.concurrent.ThreadPoolExecutor;
import java.util.concurrent.TimeUnit;

/**
 * Proporciona 3 alternativas para mostrar los números perfectos hasta un tope dado.
 * De utilidad para ver cómo el planteamiento de algoritmos adecuados es importante,
 * y para ver las posibilidades que da Java para realizar cómputos complejos.
 * 
 * @author german
 */
public class PerfectosConMersenne {
    //Anulamos la posibilidad de construir objetos. Es una maleta.
    private PerfectosConMersenne(){}
    
    //----------   METODO BUENO    -------------------------------------------

   /**
     * Calcula perfectos mediante su expresión basada en primos de Mersenne (2<sup>&rho;-1</sup>)*(2<sup>&rho;</sup>-1). 
     * 
     * @param maxRo valor máximo de "p" en la fórmula a calcular 
     */
    public static void perfectosBueno(int maxRo){
        Primes.PrimeGenerator pg=new Primes.PrimeGenerator();
        for (long ro=pg.nextPrime(), pot2=2, mersenne, np; ro<maxRo; np=pg.nextPrime(),pot2<<=np-ro, ro=np )
            if (Primes.isPrime(mersenne=(pot2<<1)-1)) System.out.println(ro+" "+pot2*mersenne);
    } 
    
    //----------   METODO BIG    ---------------------------------------------

  /**
     * Calcula perfectos mediante su expresión basada en primos de Mersenne (2<sup>&rho;-1</sup>)*(2<sup>&rho;</sup>-1),
     * utilizando una representación numérica de precisión indefinita. 
     * OJO. El resultado no es seguro. La primalidad es probabilística.
     * 
     * @param maxRo máxima potencia (p en la fórmula)
     * @param certainty factor de probabilidad del test de primalidad según err&lt;=(1-1/2^factProb)
     */
    public static void perfectosBig(int maxRo,int certainty){
        Primes.PrimeGenerator pg=new Primes.PrimeGenerator();
        BigInteger pot2=BigInteger.TWO, mersenne;
        for (long ro=pg.nextPrime(), np; ro<=maxRo; np=pg.nextPrime(),pot2=pot2.shiftLeft((int)(np-ro)), ro=np )
            if ((mersenne=pot2.shiftLeft(1).subtract(BigInteger.ONE)  ).isProbablePrime(certainty))
                System.out.println(ro+" "+pot2.multiply(mersenne));
    }
    
    //----------   METODO BIGTHREADED    --------------------------------------
    private static long executionStartTime;

    /**
     * Calcula perfectos mediante su expresión basada en primos de Mersenne (2<sup>&rho;-1</sup>)*(2<sup>&rho;</sup>-1),
     * utilizando una representación numérica de precisión indefinida y usando todos los hilos del procesador. 
     * OJO. El resultado no es seguro. La primalidad es probabilística.
     * La evaluación de cada candidato se delega en hilos "DeterminaPerfecto" que son ejecutados por un ThreadPool.
     * 
     * @param maxRo máxima potencia (p)
     * @param certainty certeza del test de primalidad según err&lt;=(1-1/2^certainty)
     * @param numeroDeHilos numero de hilos a utilizar (aparte del main)
     */

    public static void perfectosBigThreaded(int maxRo, int certainty,int numeroDeHilos) {
        System.out.println("       t(seg.)      ro  long. perfecto");
        ThreadPoolExecutor executor =  new ThreadPoolExecutor(numeroDeHilos, numeroDeHilos, 1, TimeUnit.DAYS, new ArrayBlockingQueue<Runnable>(numeroDeHilos+4), new ThreadPoolExecutor.CallerRunsPolicy());
        executionStartTime=System.currentTimeMillis();
        Primes.PrimeGenerator pg=new Primes.PrimeGenerator();
        BigInteger pot2=BigInteger.TWO;
        for (long ro=pg.nextPrime(), np; ro<=maxRo; np=pg.nextPrime(),pot2=pot2.shiftLeft((int)(np-ro)), ro=np )
            executor.execute(new DeterminaPerfecto(ro,pot2, certainty));
        executor.shutdown();
        try {executor.awaitTermination(1, TimeUnit.DAYS);
        } catch (InterruptedException ex) {System.err.println("Oops! "+ex.getMessage());
        }
    }
   
    private static class DeterminaPerfecto implements Runnable {
        private final long ro;
        private final int certainty;
        private final BigInteger mersenne, pot2;
        
        public DeterminaPerfecto(long ro, BigInteger pot2, int confidence) {
            this.ro=ro;
            this.pot2=pot2;
            this.mersenne=pot2.shiftLeft(1).subtract(BigInteger.ONE);
            this.certainty=confidence;
        }
        
        @Override
        public void run() {
            Thread.currentThread().setPriority(4);
            if (mersenne.isProbablePrime(certainty)) {
                String perfecto=pot2.multiply(mersenne).toString();
                System.out.printf("%10.3f seg. %6d %6d %s\n",(System.currentTimeMillis()-executionStartTime)/1000.0,ro,perfecto.length(),perfecto);
            }
        }
    }  
}

