package edu.upvehu.gbg.docencia.perfectos;

import java.util.ArrayList;
import java.util.List;

/**
 * Clase que proporciona métodos para generar y verificar números primos.
 *
 * <p>Esta clase incluye una lista estática de números primos precomputados que se amplía bajo demanda,
 * así como métodos para añadir nuevos primos a la lista, realizar búsquedas binarias en la lista,
 * generar secuencias de números primos y verificar si un número dado es primo.</p>
 * 
 * @version 0.1 Beta (20 dic. 2024)
 * @author german.bordel@ehu.es
 */
public class Primes{

     /**
     * Lista estática de números primos precomputados que crecerá bajo demanda.
     */
    private static final List<Long> primes=new ArrayList<>();

    /**
     * Bloque estático que inicializa la lista de números primos con los valores estrictamente necesarios.
     */
    static {
        primes.add(2L);
        primes.add(3L);
    }
    
    /**
     * Añade el siguiente número primo a la lista de primos precomputados.
     *
     * <p>Detalles de la implementación:</p>
     * <ul>
     *   <li>Prueba cada número impar como candidato a primo, comenzando con el siguiente número impar después del último primo en la lista.</li>
     *   <li>Si el candidato no es divisible por ningún primo menor o igual a su raíz cuadrada, se añade a la lista.</li>
     * </ul>
     */ 
    private static void addNextToList() {
        nextCandidate:
        for (long p=primes.get(primes.size()-1)+2L;;p+=2L){
            int i=0;
            long divisor=primes.get(++i);
            while (p>=divisor*divisor) {
                if (p % divisor == 0) continue nextCandidate;
                divisor=primes.get(++i);
            }
            primes.add(p);            
            return;
            }
    }
    
    /**
     * Realiza una búsqueda binaria en la lista de primos precomputados para encontrar un valor específico.
     *
     * <p>Detalles de la implementación:</p>
     * <ul>
     *   <li>Si el valor se encuentra, devuelve el índice del valor en la lista más 1.</li>
     *   <li>Si el valor no se encuentra, devuelve el índice del mayor primo menor que el valor, en negativo.</li>
     * </ul>
     *
     * @param n El valor a buscar en la lista de primos.
     * @return El índice del valor en la lista más 1 si se encuentra, o el índice del mayor primo menor que el valor en negativo si no se encuentra.
     */
    private static int binarySearchPrime(long n) {
        int left = 0, right = primes.size() - 1;
        while (left <= right) {
            int mid = left + (right - left) / 2;
            long midValue = primes.get(mid);

            if (midValue < n) left = mid + 1;
            else if (midValue > n) right = mid - 1;
            else return mid+1;
        }
        // No se encontró --> se devuelve el índice del mayor primo menor que value, en negativo 
        return -left; 
    }

    /**
     * Generador de una secuencia de números primos.
     *
     * <p>Esta clase proporciona métodos para generar números primos de manera secuencial,
     * utilizando una lista
     *de números primos precomputados para optimizar el proceso.</p>
     */
    public static class PrimeGenerator {

        /**
         * Índice en la lista
         *de primos precomputada al próximo primo a dar con {@link #nextPrime()}
         * @see #nextPrime()
         */
        private int index; //indice el próximo primo a dar

        /**
         * Constructor por defecto para un generador que comenzará por proporcionar el 2 al llamar a {@link #nextPrime()}.
         *  
         * <p>Detalle de la implementación:<br>
         * La función situa el índice a 0 sobre una lista
         *de números primos precomputados.</p>
         */
        public PrimeGenerator() { index=0; }
        
        /**
         * Constructor para un generador que comenzará por proporcionar el primer primo mayor que el parámetro <code>from</code> al llamar a {@link #nextPrime()}.
         *
         * <p>Detalle de la implementación:<br>
         * La función ajusta el índice sobre una lista
         *de números primos precomputados.</p>
         * <ul>
         *   <li>Si el último primo en la lista
         *es menor que el número dado, se añaden primos hasta que se supere dicho número.</li>
         *   <li>Si el número dado está en el rango de la lista
         *, realiza una búsqueda para ajustar el índice a la posición adecuada, añadiendo un primo si es necesario</li>
         * </ul>
         *
         * @param from El número a partir del cual se inicializa el índice.
         */
        public PrimeGenerator(long from) {
            index=primes.size()-1;
            if (primes.get(index)<from) {
                do addNextToList(); while (primes.get(++index)<=from);
                return;
            }
            index = binarySearchPrime(from);
            if (index<0) index = -index;
            if (index==primes.size()-1) addNextToList();
            index++;
            index = Math.abs(binarySearchPrime(from));
        }
        
        /**
         * Devuelve el siguiente número primo en la secuencia.
         *
         * <p>Detalles de la implementación:<br>
         * La función utiliza una lista
         *de números primos precomputados para optimizar la generación.</p>
         * <ul>
         *   <li>Si se ha agotado la lista
         *, se añade un primo más.</li>
         *   <li>Devuelve el primo de la lista
         *correspondiente al índice actual y avanza el índice para la siguiente llamada.</li>
         * </ul>
         *
         * @return El siguiente número primo.
         */
        public long nextPrime(){
            //Si se ha agotado la lista, añadimos un primo más buscando entre todos los impares siguientes
            if (index>primes.size()-1) addNextToList();
            //Devolvemos el primo de la lista que toca por indice y se avanza este para la siguiente llamada
            return primes.get(index++);
        }
    }
    
    /**
     * Verifica si un número dado es primo.
     *
     * <p>Detalles de la implementación:<br>
     * La función utiliza una lista
     *de números primos precomputados para optimizar la verificación</p>
     * <ul>
     *   <li>Si el número es menor que 2, se considera no primo.</li>
     *   <li>Si el número está en el rando de la lista
     *de primos, se verifica mediante búsqueda binaria.</li>
     *   <li>Si el número no está en dicho rango, se prueba con los primos de la lista
     *como divisores.</li>
     *   <li>Si no son suficientes los divisores en la lista
     *, se amplía la lista
     *elemento a elemento y se continúa la verificación.</li>
     * </ul>
     * 
     * @param n El número a verificar.
     * @return true si el número es primo, false en caso contrario.
     */
    public static boolean isPrime(long n) {
        if (n<2L) return false; 
        //Si está en el rango cubierto por la lista, nos basamos en ella
        if (n<=primes.get(primes.size()-1)) if (binarySearchPrime(n)>=0) return true;
        //Mientras disponga de primos en la lista, se prueban como divisores
        for (int index=0;index<primes.size(); index++) {
            long divisor=primes.get(index);
            if (divisor*divisor>n) return true;
            if (n%divisor==0L) return false;
        }
        //Si hay que seguir probando con divisores más allá de la lista, hacemos crecer la lista
       
        int i=primes.size();
        do{
            addNextToList();
            long divisor=primes.get(i++);
            if (divisor*divisor>n) return true;
            if (n%divisor==0L) return false;
        } while (true);
   }

}


